// Author: Ivan Kazmenko (gassa@mail.ru)
module puzzle;
// version = doMultiTile;
import board;
import button;
import common;
import multi_tile;
import rect;
import tile;

import allegro5.allegro;
import allegro5.allegro_audio;
import allegro5.allegro_image;

import core.thread;
import std.algorithm;
import std.conv;
import std.exception;
import std.random;
import std.range;
import std.string;

class Puzzle : Rect
{
	int ph;
	int pw;
	ALLEGRO_BITMAP * fullImage;
	ALLEGRO_SAMPLE * [] sample;
	ALLEGRO_SAMPLE_INSTANCE * [] [] tileSound;
	float piecePre;
	float pieceSeconds;
	int lenFreq;
	Board board;

	this (string [] level)
	{
		super (null, 0, 0, maxX, maxY,
		    0, al_map_rgb_f (0.0, 0.0, 0.0));
		ph = level[0].strip.to !(int);
		pw = level[1].strip.to !(int);
		auto imageName = "data/" ~ level[2].strip;
		fullImage = al_load_bitmap (imageName.ptr);
		sample = new ALLEGRO_SAMPLE * [ph];
		piecePre = level[3].strip.to !(float);
		pieceSeconds = level[4].strip.to !(float);
		foreach (row; 0..ph)
		{
			import std.stdio;
			// bug? does not work without the dot
			auto curName = "./data/" ~ level[5 + row];
//			writeln (curName);
			sample[row] = al_load_sample (curName.ptr);
//			writeln (sample[row]);
//			writeln (al_get_sample_length (sample[row]));
		}
/*
		ALLEGRO_SAMPLE_ID y;
		al_play_sample (sample.front, 1.0, 0.0, 1.0,
		    ALLEGRO_PLAYMODE.ALLEGRO_PLAYMODE_ONCE, &y);
*/
		tileSound = new ALLEGRO_SAMPLE_INSTANCE * [] [] (ph, pw);
		lenFreq = cast (int) (pieceSeconds * commonFreq);
		foreach (row; 0..ph)
		{
			foreach (col; 0..pw)
			{
				tileSound[row][col] =
				    al_create_sample_instance (sample[row]);
/*
				al_set_sample_instance_position
				    (tileSound[row][col], cast (int)
				    (commonFreq *
				    (piecePre + pieceSeconds * col)));
*/
				enforce (al_attach_sample_instance_to_mixer
				    (tileSound[row][col], mixer));
/*
				al_set_sample_instance_length
				    (tileSound[row][col], lenFreq);
				al_set_sample_instance_gain
				    (tileSound[row][col], 1.0);
				al_set_sample_instance_speed
				    (tileSound[row][col], 1.0);
				al_set_sample_instance_pan
				    (tileSound[row][col], 0.0);
*/
				al_set_sample_instance_playmode
				    (tileSound[row][col],
				    ALLEGRO_PLAYMODE.ALLEGRO_PLAYMODE_ONCE);
/*
				al_set_sample_instance_playing
				    (tileSound[row][col], true);
*/
			}
		}
		import std.stdio;
//		writeln (al_get_sample_instance_length (tileSound[1][1]));
//		al_play_sample_instance (tileSound[1][1]);
//		writeln (al_get_sample_instance_playing (tileSound[1][1]));
		auto levelMap = level[5 + ph..$]
		    .map !(strip).map !(split).array;
		auto tw = 80;
		auto th = 80;
		auto bx = 100;
		auto by = 100;
		auto tile = new Tile [] [] (ph, pw);

		version (doMultiTile)
		{
			Tile [] [string] multi;
			foreach (row; 0..ph)
			{
				foreach (col; 0..pw)
				{
					tile[row][col] = new Tile (null, 0, 0,
					    tw, th, 0,
					    al_map_rgba_f (0.0, 0.0, 0.0, 1.0),
					    this, row, col,
					    al_create_sub_bitmap
					    (fullImage, col * tw, row * th, tw, th));
					multi[levelMap[row][col]] ~=
					    tile[row][col];
				}
			}
			foreach (list; multi.byValue)
			{
				auto rowLo = list.map
				    !(cur => cur.row).minElement;
				auto rowHi = list.map
				    !(cur => cur.row).maxElement;
				auto colLo = list.map
				    !(cur => cur.col).minElement;
				auto colHi = list.map
				    !(cur => cur.col).maxElement;
				auto tempW = (colHi - colLo + 1) * tw;
				auto tempH = (rowHi - rowLo + 1) * th;
				auto rx = uniform !("[]")
				    (bx, bx + tw * pw - tempW);
				auto ry = uniform !("[]")
				    (by + th * ph, maxY - 60 * 2 - tempH);
				auto temp = new MultiTile (this,
				    rx, ry, tempW, tempH);
				foreach (cur; list)
				{
					cur.parent = temp;
					temp.child ~= cur;
					cur.relX = (cur.col - colLo) * tw;
					cur.relY = (cur.row - rowLo) * th;
				}
			}
		}
		else
		{
			foreach (row; 0..ph)
			{
				foreach (col; 0..pw)
				{
					auto rx = uniform !("[]")
					    (bx, bx + tw * pw - tw);
					auto ry = uniform !("[]")
					    (by + th * ph, maxY - 60 * 2 - th);
					tile[row][col] = new Tile (this,
					    rx, ry, tw, th, 0,
					    al_map_rgba_f (0.0, 0.0, 0.0, 1.0),
					    this, row, col,
					    al_create_sub_bitmap
					    (fullImage, col * tw, row * th,
					    tw, th));
				}
			}
		}
    		board = new Board (this, bx, by, tw * pw, th * ph,
		    0, al_map_rgb_f (0.2, 0.2, 0.2), pw, ph, tw, th,
		    levelMap, tile);
		auto buttonColor = al_map_rgb_f (0.1, 0.3, 0.5);
		auto exitButton = new Button (this,
		    (maxX - 140) / 2, maxY - 60, 140, 40, 5,
		    buttonColor, al_map_rgb_f (0.9, 0.5, 0.5),
		    buttonFont, "Quit".toAllegroUstr (),
		    (int posX, int posY)
		    {screenRoot = menu; cursorRoot = null;});
	}

	~this ()
	{
		al_destroy_bitmap (fullImage);
	}

	override bool unclickThis (int posX, int posY)
	{
		if (cursorRoot is null)
		{
			return false;
		}
		cursorRoot.relX = posX - cursorRoot.w / 2;
		cursorRoot.relY = posY - cursorRoot.h / 2;
		child ~= cursorRoot;
		cursorRoot.parent = this;
		cursorRoot = null;
		return true;
	}

	void playPart (int row, int col)
	{
		import std.stdio;
//		writeln ("?");
//		writeln (al_get_voice_playing (voice));
//		enforce (al_set_voice_playing (voice, false));
/*
		foreach (curRow; 0..ph)
		{
			foreach (curCol; 0..pw)
			{
				enforce (al_detach_sample_instance
				    (tileSound[curRow][curCol]));
			}
		}
		enforce (al_attach_sample_instance_to_mixer
		    (tileSound[row][col], mixer));
*/
/*
		al_set_sample_instance_position
		    (tileSound[row][col], lenFreq * col);
		al_set_sample_instance_length
		    (tileSound[row][col], lenFreq);
*/
		al_stop_sample_instance (tileSound[row][col]);
		al_set_sample_instance_position
		    (tileSound[row][col], cast (int) (commonFreq *
		    (piecePre + pieceSeconds * col)));
		al_play_sample_instance (tileSound[row][col]);
		new Thread (()
		{
			al_rest (pieceSeconds);
			al_stop_sample_instance (tileSound[row][col]);
			import std.stdio;
//			writeln ("!");
		}).start ();
/*
		auto t = al_create_timer (pieceSeconds);
		al_register_event_source (al_get_timer_event_source (t));
		al_start_timer (t);
*/
//		enforce (al_set_voice_playing (voice, true));
	}
}
