// Author: Ivan Kazmenko (gassa@mail.ru)
module quartet;
import button;
import common;
import obj;
import puzzle;
import text_rect;
import rect;

import core.stdc.stdlib;
import std.algorithm;
import std.exception;
import std.range;
import std.stdio;

import allegro5.allegro;
import allegro5.allegro_acodec;
import allegro5.allegro_audio;
import allegro5.allegro_font;
import allegro5.allegro_image;
import allegro5.allegro_primitives;
import allegro5.allegro_ttf;

void init ()
{
	enforce (al_init ());
	enforce (al_install_audio ());
	enforce (al_install_keyboard ());
	enforce (al_install_mouse ());
	enforce (al_init_acodec_addon ());
	enforce (al_init_font_addon ());
	enforce (al_init_image_addon ());
	enforce (al_init_primitives_addon ());
	enforce (al_init_ttf_addon ());

	display = al_create_display (maxX, maxY);
	enforce (display);

	drawTimer = al_create_timer (1.0 / fps);
	enforce (drawTimer);
	al_start_timer (drawTimer);

	textFont = al_load_ttf_font ("data/EBGaramond.otf", 24, 0);
	enforce (textFont);

	buttonFont = al_load_ttf_font ("data/EBGaramond.otf", 30, 0);
	enforce (buttonFont);

	captionFont = al_load_ttf_font ("data/EBGaramond.otf", 48, 0);
	enforce (captionFont);

	eventQueue = al_create_event_queue ();
	enforce (eventQueue);

	al_register_event_source (eventQueue,
	    al_get_keyboard_event_source ());
	al_register_event_source (eventQueue,
	    al_get_mouse_event_source ());
	al_register_event_source (eventQueue,
	    al_get_timer_event_source (drawTimer));
	al_register_event_source (eventQueue,
	    al_get_display_event_source (display));

	mixer = al_create_mixer (commonFreq,
	    ALLEGRO_AUDIO_DEPTH.ALLEGRO_AUDIO_DEPTH_FLOAT32,
	    ALLEGRO_CHANNEL_CONF.ALLEGRO_CHANNEL_CONF_2);
	enforce (mixer);
	voice = al_create_voice (commonFreq,
	    ALLEGRO_AUDIO_DEPTH.ALLEGRO_AUDIO_DEPTH_INT16,
	    ALLEGRO_CHANNEL_CONF.ALLEGRO_CHANNEL_CONF_2);
	enforce (voice);
	import std.stdio;
//	writeln (al_get_voice_playing (voice));
	enforce (al_attach_mixer_to_voice (mixer, voice));
//	writeln (al_get_voice_playing (voice));
	enforce (al_reserve_samples (64));
//	enforce (al_set_voice_playing (voice, false));
/*
	auto x = al_load_sample ("data/1-1.ogg");
	writeln (x);
	ALLEGRO_SAMPLE_ID y;
	al_play_sample (x, 1.0, 0.0, 1.0,
	    ALLEGRO_PLAYMODE.ALLEGRO_PLAYMODE_ONCE, &y);
*/
}

void normalizeMouseState (MouseState) (ref MouseState mouse)
{
	if (cursorRoot is null)
	{
		return;
	}
	auto mouseCopy = mouse;
	mouse.x = max (mouse.x, cursorRoot.relX);
	mouse.x = min (mouse.x, maxX - cursorRoot.w + cursorRoot.relX);
	mouse.y = max (mouse.y, cursorRoot.relY);
	mouse.y = min (mouse.y, maxY - cursorRoot.h + cursorRoot.relY);
	if (mouse != mouseCopy)
	{
		al_set_mouse_xy (display, mouse.x, mouse.y);
	}
}

void draw ()
{
	al_clear_to_color (al_map_rgb_f (0.0, 0.0, 0.0));
	enforce (screenRoot !is null);
	screenRoot.draw ();
	ALLEGRO_MOUSE_STATE mouse;
	al_get_mouse_state (&mouse);
	if (cursorRoot !is null)
	{
		normalizeMouseState (mouse);
		ALLEGRO_BITMAP * prevBuffer = al_get_target_bitmap ();
		ALLEGRO_BITMAP * curBuffer = al_create_sub_bitmap
		    (prevBuffer, mouse.x - cursorRoot.relX,
		    mouse.y - cursorRoot.relY, cursorRoot.w, cursorRoot.h);
		al_set_target_bitmap (curBuffer);
		int sRelX = 0;
		int sRelY = 0;
		swap (sRelX, cursorRoot.relX);
		swap (sRelY, cursorRoot.relY);
		cursorRoot.draw ();
		swap (sRelX, cursorRoot.relX);
		swap (sRelY, cursorRoot.relY);
		al_set_target_bitmap (prevBuffer);
		al_destroy_bitmap (curBuffer);
	}
	al_flip_display ();
}

bool isFinished = false;

auto prepareMenu ()
{
	auto menu = new Rect (null, 0, 0, maxX, maxY, 0,
	    al_map_rgb_f (0.0, 0.0, 0.0));
	auto buttonColor = al_map_rgb_f (0.1, 0.3, 0.5);
	auto caption = new TextRect (menu,
	    (maxX - 300) / 2, 115, 300, 60, 0,
	    al_map_rgba_f (0.0, 0.0, 0.0, 0.0), al_map_rgb_f (0.4, 0.9, 0.1),
	    captionFont, "QUARTET".toAllegroUstr ());
	auto someText = new TextRect (menu,
	    (maxX - 400) / 2, 190, 400, 50, 0,
	    al_map_rgba_f (0.0, 0.0, 0.0, 0.0), al_map_rgb_f (0.5, 0.8, 0.8),
	    textFont, "Assemble the pieces".toAllegroUstr ());
	auto level1 = new Button (menu,
	    maxX * 1 / 4 - 140 / 2, 250, 140, 40, 5,
	    buttonColor, al_map_rgb_f (0.9, 0.9, 0.5),
	    buttonFont, "Level 1".toAllegroUstr (),
	    (int posX, int posY)
	    {screenRoot = preparePuzzle ("data/1.txt");});
/*
	auto levelTest = new Button (menu,
	    maxX * 1 / 4 - 140 / 2, 300, 140, 40, 5,
	    buttonColor, al_map_rgb_f (0.9, 0.9, 0.5),
	    buttonFont, "Test Puzzle".toAllegroUstr (),
	    (int posX, int posY)
	    {screenRoot = preparePuzzle ("data/test.txt");});
*/
	auto exitButton = new Button (menu,
	    (maxX - 140) / 2, 375, 140, 40, 5,
	    buttonColor, al_map_rgb_f (0.9, 0.5, 0.5),
	    buttonFont, "Exit".toAllegroUstr (),
	    (int posX, int posY) {isFinished = true;});
	return menu;
}

auto preparePuzzle (string fileName)
{
	return new Puzzle (File (fileName, "rt").byLineCopy.array);
}

void mainLoop ()
{
	menu = prepareMenu ();
	screenRoot = menu;
	cursorRoot = null;
	draw ();

	isFinished = false;
	while (!isFinished)
	{
		ALLEGRO_EVENT currentEvent;
		al_wait_for_event (eventQueue, &currentEvent);

		switch (currentEvent.type)
		{
		case ALLEGRO_EVENT_DISPLAY_CLOSE:
			isFinished = true;
			break;

		case ALLEGRO_EVENT_MOUSE_BUTTON_DOWN:
			if (currentEvent.mouse.button == 1)
			{
				normalizeMouseState (currentEvent.mouse);
				auto x = currentEvent.mouse.x;
				auto y = currentEvent.mouse.y;
				screenRoot.click (x, y);
			}
			break;

		case ALLEGRO_EVENT_MOUSE_BUTTON_UP:
			if (currentEvent.mouse.button == 1)
			{
				normalizeMouseState (currentEvent.mouse);
				auto x = currentEvent.mouse.x;
				auto y = currentEvent.mouse.y;
				screenRoot.unclick (x, y);
			}
			break;

		case ALLEGRO_EVENT_TIMER:
			draw ();
			break;

		default:
			break;
		}
	}
}

void happyEnd ()
{
	al_detach_mixer (mixer);
	al_destroy_voice (voice);
	al_destroy_mixer (mixer);

	al_destroy_event_queue (eventQueue);
	al_destroy_font (captionFont);
	al_destroy_font (buttonFont);
	al_destroy_font (textFont);
	al_destroy_timer (drawTimer);
	al_destroy_display (display);

	al_shutdown_font_addon ();
	al_shutdown_image_addon ();
	al_shutdown_primitives_addon ();
	al_shutdown_ttf_addon ();
	al_uninstall_audio ();

	exit (EXIT_SUCCESS);
}

int main (string [] args)
{
	return al_run_allegro (
	{
		init ();
		mainLoop ();
		happyEnd ();
		return 0;
	});
}
